#-----------------------------------------------------------------------------#
#   base.py                                                                   #
#                                                                             #
#   Copyright © 2015-2017, Rajiv Bakulesh Shah, original author.              #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



import abc
import contextlib
import functools
import json
import os
import random
import string

from redis import Redis
from redis import WatchError

from . import monkey
from .exceptions import RandomKeyError
from .exceptions import TooManyTriesError



class _Common:
    _DEFAULT_REDIS_URL = 'http://localhost:6379/'
    _NUM_TRIES = 3
    _RANDOM_KEY_PREFIX = 'pottery:'
    _RANDOM_KEY_LENGTH = 16

    @staticmethod
    def _encode(value):
        encoded = json.dumps(value, sort_keys=True)
        return encoded

    @staticmethod
    def _decode(value):
        decoded = json.loads(value.decode('utf-8'))
        return decoded

    def __init__(self, *args, redis=None, key=None, **kwargs):
        self.redis = redis
        self.key = key

    def __del__(self):
        if self.key.startswith(self._RANDOM_KEY_PREFIX):
            self.redis.delete(self.key)

    def __eq__(self, other):
        if type(self) is type(other) and \
           self.redis == other.redis and \
           self.key == other.key:
            equals = True
        else:
            equals = super().__eq__(other)
            if equals is NotImplemented:
                equals = False
        return equals

    def __ne__(self, other):
        does_not_equal = not self.__eq__(other)
        return does_not_equal

    @property
    def _default_redis(self):
        url = os.environ.get('REDIS_URL', self._DEFAULT_REDIS_URL)
        redis = Redis.from_url(url)
        return redis

    @property
    def redis(self):
        return self._redis

    @redis.setter
    def redis(self, value):
        self._redis = self._default_redis if value is None else value

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = value or self._random_key()

    def _random_key(self, *, tries=_NUM_TRIES):
        if tries <= 0:
            raise RandomKeyError(self.redis, self.key)
        all_chars = string.digits + string.ascii_letters
        random_char = functools.partial(random.choice, all_chars)
        suffix = ''.join(random_char() for n in range(self._RANDOM_KEY_LENGTH))
        random_key = self._RANDOM_KEY_PREFIX + suffix
        if self.redis.exists(random_key):
            random_key = self._random_key(tries=tries-1)
        return random_key



class Pipelined(metaclass=abc.ABCMeta):
    @abc.abstractproperty
    def _NUM_TRIES(self):
        'The number of times to try generating a random key before giving up.'

    @abc.abstractproperty
    def redis(self):
        'Redis client.'

    @abc.abstractproperty
    def key(self):
        'Redis key.'

    @property
    @contextlib.contextmanager
    def _pipeline(self):
        pipeline = self.redis.pipeline()
        try:
            yield pipeline
        finally:
            pipeline.execute()

    def _watch(func):
        @functools.wraps(func)
        def wrap(self, *args, **kwargs):
            for _ in range(self._NUM_TRIES):
                try:
                    original_redis = self.redis
                    with self._pipeline as pipeline:
                        self.redis = pipeline
                        self.redis.watch(self.key)
                        value = func(self, *args, **kwargs)
                    return value
                except WatchError:
                    pass
                finally:
                    self.redis = original_redis
            else:
                raise TooManyTriesError(self.redis, self.key)
        return wrap



class _Clearable(metaclass=abc.ABCMeta):
    @abc.abstractproperty
    def redis(self):
        'Redis client.'

    @abc.abstractproperty
    def key(self):
        'Redis key.'

    def clear(self):
        'Remove the elements in a Redis-backed container.  O(n)'
        self.redis.delete(self.key)



class Base(_Common, _Clearable, Pipelined):
    ...



class Iterable(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _decode(value):     # pragma: no cover
        ...

    @abc.abstractproperty   # pragma: no cover
    def key(self):
        'Redis key.'

    @abc.abstractmethod     # pragma: no cover
    def _scan(self, key, *, cursor=0):
        ...

    def __iter__(self):
        'Iterate over the items in a Redis-backed container.  O(n)'
        cursor = 0
        while True:
            cursor, iterable = self._scan(self.key, cursor=cursor)
            for value in iterable:
                decoded = self._decode(value)
                yield decoded
            if cursor == 0:
                break



class Primitive(metaclass=abc.ABCMeta):
    _default_masters = frozenset({Redis()})

    def __init__(self, *, key, masters=_default_masters):
        self.key = key
        self.masters = masters or self._default_masters

    @abc.abstractproperty
    def KEY_PREFIX(self):
        'Redis key prefix/namespace.'

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = '{}:{}'.format(self.KEY_PREFIX, value)



#-----------------------------------------------------------------------------#
#   list.py                                                                   #
#                                                                             #
#   Copyright © 2015-2017, Rajiv Bakulesh Shah, original author.              #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



import collections.abc
import functools
import itertools

from redis import ResponseError

from .base import Base
from .base import Pipelined
from .exceptions import KeyExistsError



class RedisList(Base, collections.abc.MutableSequence):
    'Redis-backed container compatible with Python lists.'

    def _raise_on_error(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ResponseError:
                raise IndexError('list assignment index out of range')
        return wrap

    def _slice_to_indices(self, slice_or_index):
        try:
            start = slice_or_index.start or 0
            stop = slice_or_index.stop or len(self)
            step = slice_or_index.step or 1
            indices = range(start, stop, step)
        except AttributeError:
            indices = (slice_or_index,)
        return indices

    def __init__(self, iterable=tuple(), *, redis=None, key=None):
        'Initialize a RedisList.  O(1)'
        super().__init__(iterable, redis=redis, key=key)
        self._populate(iterable)

    @Pipelined._watch
    def _populate(self, iterable=tuple()):
        values = [self._encode(value) for value in iterable]
        if values:
            if self.redis.exists(self.key):
                raise KeyExistsError(self.redis, self.key)
            else:
                self.redis.multi()
                self.redis.rpush(self.key, *values)

    # Methods required by collections.abc.MutableSequence:

    def __getitem__(self, index):
        'l.__getitem__(index) <==> l[index].  O(n)'
        try:
            value = self.redis.lindex(self.key, index)
            if value is None:
                raise IndexError('list index out of range')
            else:
                decoded = self._decode(value)
        except ResponseError:
            # This is monumentally stupid.  Python's list API requires us to
            # get elements by slice (defined as a start index, a stop index,
            # and a step).  Of course, Redis allows us to only get elements by
            # start and stop (no step), because it's Redis.  So our ridiculous
            # hack is to get all of the elements between start and stop from
            # Redis, then discard the ones between step in Python.  More info:
            # http://redis.io/commands/lrange
            indices = self._slice_to_indices(index)
            values = self.redis.lrange(self.key, indices[0], indices[-1])
            values = values[::index.step]
            decoded = [self._decode(value) for value in values]
        return decoded

    @_raise_on_error
    def __setitem__(self, index, value):
        'l.__setitem__(index, value) <==> l[index] = value.  O(n)'
        try:
            self.redis.lset(self.key, index, self._encode(value))
        except ResponseError:
            with self._pipeline as pipeline:
                indices, values = self._slice_to_indices(index), value
                values = [self._encode(value) for value in values]
                for index, value in zip(indices, values):
                    pipeline.lset(self.key, index, value)
                indices, num = indices[len(values):], 0
                for index in indices:
                    pipeline.lset(self.key, index, None)
                    num += 1
                if num:
                    pipeline.lrem(self.key, None, num=num)

    @_raise_on_error
    def __delitem__(self, index):
        'l.__delitem__(index) <==> del l[index].  O(n)'
        # This is monumentally stupid.  Python's list API requires us to
        # delete an element by *index.*  Of course, Redis doesn't support
        # that, because it's Redis.  Instead, Redis supports deleting an
        # element by *value.*  So our ridiculous hack is to set l[index] to
        # None, then to delete the value None.  More info:
        # http://redis.io/commands/lrem
        with self._pipeline as pipeline:
            indices, num = self._slice_to_indices(index), 0
            for index in indices:
                pipeline.lset(self.key, index, None)
                num += 1
            if num: # pragma: no cover
                pipeline.lrem(self.key, None, num=num)

    def __len__(self):
        'Return the number of items in a RedisList.  O(1)'
        return self.redis.llen(self.key)

    @Pipelined._watch
    def insert(self, index, value):
        'Insert an element into a RedisList before the given index.  O(n)'
        value = self._encode(value)
        if index <= 0:
            self.redis.multi()
            self.redis.lpush(self.key, value)
        elif index < len(self):
            # This is monumentally stupid.  Python's list API requires us to
            # insert an element before the given *index.*  Of course, Redis
            # doesn't support that, because it's Redis.  Instead, Redis
            # supports inserting an element before a given (pivot) *value.*  So
            # our ridiculous hack is to set the pivot value to None, then to
            # insert the desired value and the original pivot value before the
            # value None, then to delete the value None.  More info:
            # http://redis.io/commands/linsert
            pivot = self._encode(self[index])
            self.redis.multi()
            self.redis.lset(self.key, index, None)
            for value in (value, pivot):
                self.redis.linsert(self.key, 'BEFORE', None, value)
            self.redis.lrem(self.key, None, num=1)
        else:
            self.redis.multi()
            self.redis.rpush(self.key, value)

    # Methods required for Raj's sanity:

    def sort(self, *, key=None, reverse=False):
        'Sort a RedisList in place.  O(n)'
        if key is None:
            self.redis.sort(self.key, desc=reverse, store=self.key)
        else:
            raise NotImplementedError('sorting by key not implemented')

    def __add__(self, other):
        'Append the items in other to a RedisList.  O(1)'
        iterable = itertools.chain(self, other)
        return self.__class__(iterable, redis=self.redis)

    def __repr__(self):
        'Return the string representation of a RedisList.  O(n)'
        list_ = self.redis.lrange(self.key, 0, -1)
        list_ = [self._decode(value) for value in list_]
        return self.__class__.__name__ + str(list_)



#-----------------------------------------------------------------------------#
#   exceptions.py                                                             #
#                                                                             #
#   Copyright © 2015-2017, Rajiv Bakulesh Shah, original author.              #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



class PotteryError(Exception):
    'Base exception class.'

    def __init__(self, redis, key):
        self._redis = redis
        self._key = key

    def __str__(self):
        return str((self._redis, self._key))



class KeyExistsError(PotteryError):
    'Initializing a container on a Redis key that already exists.'

    def __str__(self):
        return self._key



class RandomKeyError(PotteryError):
    "Can't create a random Redis key; all of our attempts already exist."

    def __str__(self):
        return str(self._redis)



class TooManyTriesError(PotteryError):
    "Can't complete a Redis transaction; tried too many times."
